/**
 * Copyright (C) 2010-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.terremark.compute;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageFormat;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.MachineImageSupport;
import org.dasein.cloud.compute.MachineImageType;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.terremark.Terremark;
import org.jclouds.rest.RestContext;
import org.jclouds.trmk.vcloud_0_8.domain.CatalogItem;
import org.jclouds.trmk.vcloud_0_8.domain.ReferenceType;
import org.jclouds.trmk.vcloud_0_8.domain.VAppTemplate;
import org.jclouds.trmk.vcloudexpress.TerremarkVCloudExpressAsyncClient;
import org.jclouds.trmk.vcloudexpress.TerremarkVCloudExpressClient;

import javax.annotation.Nonnull;

public class Template implements MachineImageSupport {
    static URI getCatalogUri(RestContext<TerremarkVCloudExpressClient, TerremarkVCloudExpressAsyncClient> ctx, String id) {
        if( id.startsWith("http") ) {
            return URI.create(id);
        }
        return URI.create(ctx.getEndpoint() + "/v" + ctx.getApiVersion() + "/catalogItem/" + id);            
    }
    
    static URI getTemplateUri(RestContext<TerremarkVCloudExpressClient, TerremarkVCloudExpressAsyncClient> ctx, String id) {
        if( id.startsWith("http") ) {
            return URI.create(id);
        }
        return URI.create(ctx.getEndpoint() + "/v" + ctx.getApiVersion() + "/vappTemplate/" + id);            
    }
    
    private Terremark provider;
    
    Template(Terremark provider) {
        this.provider = provider;
    }

    @Override
    public void remove(String templateId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    @Override
    public Iterable<MachineImage> searchMachineImages(String keyword, Platform platform, Architecture architecture) throws InternalException, CloudException {
        ArrayList<MachineImage> matches = new ArrayList<MachineImage>();
        
        for( MachineImage image : listMachineImages() ) {
            if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                continue;
            }
            if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                Platform mine = image.getPlatform();
                
                if( platform.isWindows() && !mine.isWindows() ) {
                    continue;
                }
                if( platform.isUnix() && !mine.isUnix() ) {
                    continue;
                }
                if( platform.isBsd() && !mine.isBsd() ) {
                    continue;
                }
                if( platform.isLinux() && !mine.isLinux() ) {
                    continue;
                }
                if( platform.equals(Platform.UNIX) ) {
                    if( !mine.isUnix() ) {
                        continue;
                    }
                }
                else if( !platform.equals(mine) ) {
                    continue;
                }
            }
            if( keyword != null && !keyword.equals("") ) {
                keyword = keyword.toLowerCase();
                if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                    if( !image.getName().toLowerCase().contains(keyword) ) {
                        if( !image.getDescription().toLowerCase().contains(keyword) ) {
                            continue;
                        }
                    }
                }
            }
            matches.add(image);
        }
        return matches;
    }

    private Architecture getArchitecture(String templateName) {
        if( templateName == null || templateName.indexOf("32-bit") != -1 || templateName.indexOf("32 bit") != -1 ) {
            return Architecture.I32;            
        }
        else if( templateName.indexOf("64-bit") != -1 || templateName.indexOf("64 bit") != -1 ) {
            return Architecture.I64;
        }
        return Architecture.I64;
    }
    
    @Override
    public MachineImage getMachineImage(String templateId) throws InternalException, CloudException {
        for( MachineImage image : listMachineImages() ) {
            if( image.getProviderMachineImageId().equals(templateId) ) {
                return image;
            }
        }
        return null;
    }

    static final Pattern LINUX_USER_PASSWORD_PATTERN = Pattern.compile(".*[Uu]sername: ([a-z]+) ?.*\n[Pp]assword: ([^ ]+) ?\n.*");

    String getPassword(String id) throws CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        VAppTemplate template = null;
                
        try {
            template = ctx.getApi().getVAppTemplate(getTemplateUri(ctx, id));
        }
        finally {
            ctx.close();
        }
        if( template == null ) {
            return "UNKNOWN";
        }
        if (template.getDescription().indexOf("Windows") >= 0) {
            return "UNKNOWN";
        } 
        else {
            Matcher matcher = LINUX_USER_PASSWORD_PATTERN.matcher(template.getDescription());
            
            if( matcher.find() ) {
                String pw = matcher.group(2);
                
                return pw;
            }
            else {
                return "UNKNOWN";
            }
        }
    }
    
    @Override
    public String getProviderTermForImage(Locale locale) {
        return "template";
    }

    @Override
    public Iterable<String> listShares(String templateId) throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachine(String serverId, String directory, String imageName) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    @Override
    public boolean supportsCustomImages() {
        return false;
    }

    @Override
    public boolean supportsImageSharing() {
        return false;
    }

    @Override
    public Iterable<MachineImage> listMachineImages() throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
        	TerremarkVCloudExpressClient client = ctx.getApi();
            ArrayList<MachineImage> images = new ArrayList<MachineImage>();
            Collection<ReferenceType> resources;
            
            
            resources = client.findCatalogInOrgNamed(null, null).values();
            for( ReferenceType rsc : resources ) {
                             
                CatalogItem catItem = client.findCatalogItemInOrgCatalogNamed(null, null, rsc.getName());
                MachineImage image  = toImage(client.getVAppTemplate(catItem.getEntity().getHref()));
                if( image != null ) {
                    images.add(image);
                }        
            }
            
            return images;
        }
        finally {
            ctx.close();
        }
    }


    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }
    
    @Override
    public void shareMachineImage(String templateId, String accountId, boolean allowed) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }
    
    private MachineImage toImage(ReferenceType rsc) {
        MachineImage image = new MachineImage();
        
        image.setCurrentState(MachineImageState.ACTIVE);
        image.setArchitecture(getArchitecture(rsc.getName()));
        image.setDescription(rsc.getName());
        image.setName(rsc.getName());
        image.setProviderOwnerId("--terremark--");
        image.setPlatform(Platform.guess(rsc.getName()));
        image.setProviderMachineImageId(Terremark.toId(rsc.getHref().toASCIIString()));
        image.setType(MachineImageType.STORAGE);
        image.setProviderRegionId(provider.getContext().getRegionId());
        image.setTags(new HashMap<String,String>());
        image.setSoftware("");
        return image;
    }

    @Override
    public void downloadImage(String machineImageId, OutputStream toOutput) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean hasPublicLibrary() {
        return false;
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public String installImageFromUpload(MachineImageFormat format, InputStream imageStream) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean isImageSharedWithPublic(String machineImageId) throws CloudException, InternalException {
        return false;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        return true;
    }

    @Override
    public Iterable<MachineImage> listMachineImagesOwnedBy(String accountId) throws CloudException, InternalException {
        if( accountId == null ) {
            return Collections.emptyList();
        }
        else if( accountId.equals(provider.getContext().getAccountNumber()) ) {
            return listMachineImages();
        }
        else {
            return Collections.emptyList();
        }
    }

    @Override
    public Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public String registerMachineImage(String atStorageLocation) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean supportsImageSharingWithPublic() {
        return false;
    }

    @Override
    public String transfer(CloudProvider fromCloud, String machineImageId) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }
}
