/**
 * Copyright (C) 2010-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.terremark.identity;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;
import java.util.Set;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.identity.SSHKeypair;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.identity.ShellKeySupport;
import org.dasein.cloud.jclouds.terremark.Terremark;
import org.jclouds.rest.RestContext;
import org.jclouds.trmk.vcloud_0_8.domain.KeyPair;
import org.jclouds.trmk.vcloud_0_8.domain.Org;
import org.jclouds.trmk.vcloudexpress.TerremarkVCloudExpressAsyncClient;
import org.jclouds.trmk.vcloudexpress.TerremarkVCloudExpressClient;

public class TerremarkKeypair implements ShellKeySupport {
	static private final Logger logger = Logger.getLogger(Terremark.class);
	
	private Terremark provider = null;
	
	public TerremarkKeypair(@Nonnull Terremark cloud) {
		this.provider =  cloud;
	}
	
	@Override
	public @Nonnull SSHKeypair createKeypair(@Nonnull String name) throws InternalException, CloudException {
		RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
		
        if( ctx == null ) {
            throw new CloudException("No context was established for this call.");
        }
        KeyPair key;
        
        try {
        	TerremarkVCloudExpressClient client = ctx.getApi();
        	Org org = client.findOrgNamed(null);
        	key = client.generateKeyPairInOrg(org.getHref(), name, false);
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
        	ctx.close();
        }
        
        SSHKeypair kp = new SSHKeypair();
        kp.setFingerprint(key.getFingerPrint());
        kp.setName(key.getName());
        try {
			kp.setPrivateKey(key.getPrivateKey().getBytes("utf-8"));
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
            throw new InternalException(e);
		}
        kp.setProviderKeypairId(Terremark.toId(String.valueOf(key.getId())));
        kp.setProviderOwnerId(provider.getContext().getAccountNumber());
        kp.setProviderRegionId(provider.getContext().getRegionId());
        kp.setPublicKey(null);
         
        return kp;
	}

	@Override
	public void deleteKeypair(@Nonnull String name) throws InternalException, CloudException {
		RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
		
        if( ctx == null ) {
            throw new CloudException("No context was established for this call.");
        }
        
        try {
        	TerremarkVCloudExpressClient client = ctx.getApi();
        	Org org = client.findOrgNamed(null);
        	KeyPair key = client.findKeyPairInOrg(org.getHref(), name);
        	client.deleteKeyPair(key.getId());
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
        	ctx.close();
        }
	}

	@Override
	public @Nullable String getFingerprint(@Nonnull String name) throws InternalException, CloudException {
		RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
		
        if( ctx == null ) {
            throw new CloudException("No context was established for this call.");
        }
        String fingerprint;
        try {
        	TerremarkVCloudExpressClient client = ctx.getApi();
        	Org org = client.findOrgNamed(null);
        	KeyPair key = client.findKeyPairInOrg(org.getHref(), name);
        	fingerprint = key.getFingerPrint();
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
        	ctx.close();
        }
        return fingerprint;
	}

    @Override
    public @Nullable SSHKeypair getKeypair(@Nonnull String name) throws InternalException, CloudException {
		RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
		
        if( ctx == null ) {
            throw new CloudException("No context was established for this call.");
        }
        KeyPair key;
        SSHKeypair kp = new SSHKeypair();
        
        try {
        	TerremarkVCloudExpressClient client = ctx.getApi();
        	Org org = client.findOrgNamed(null);
        	key = client.findKeyPairInOrg(org.getHref(), name);
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
        	ctx.close();
        }
        if (key != null){
            
            kp.setFingerprint(key.getFingerPrint());
            kp.setName(key.getName());
            try {
    			kp.setPrivateKey(key.getPrivateKey().getBytes("utf-8"));
    		} catch (UnsupportedEncodingException e) {
    			e.printStackTrace();
                throw new InternalException(e);
    		}
            kp.setProviderKeypairId(Terremark.toId(String.valueOf(key.getId())));
            kp.setProviderOwnerId(provider.getContext().getAccountNumber());
            kp.setProviderRegionId(provider.getContext().getRegionId());
            kp.setPublicKey(null);
        }
        else {
        	logger.warn("The key " + name + " could not be found!");
        	kp = null;
        }
         
        return kp;
    }

	@Override
	public @Nonnull String getProviderTermForKeypair(@Nonnull Locale locale) {
		return "key";
	}

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        provider.testContext();
        return true;
    }

    @Override
	public @Nonnull Collection<SSHKeypair> list() throws InternalException, CloudException {
		RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
		
        if( ctx == null ) {
            throw new CloudException("No context was established for this call.");
        }
        Set<KeyPair> keys;
        ArrayList<SSHKeypair> kps = new ArrayList<SSHKeypair>();
        try {
        	TerremarkVCloudExpressClient client = ctx.getApi();
        	Org org = client.findOrgNamed(null);
        	keys = client.listKeyPairsInOrg(org.getHref());
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally{
        	ctx.close();
        }
        for (KeyPair key : keys) {
            SSHKeypair kp = new SSHKeypair();
            kp.setFingerprint(key.getFingerPrint());
            kp.setName(key.getName());
            try {
    			kp.setPrivateKey(key.getPrivateKey() == null ? null : key.getPrivateKey().getBytes("utf-8"));
    		} catch (UnsupportedEncodingException e) {
    			e.printStackTrace();
                throw new InternalException(e);
    		}
            kp.setProviderKeypairId(Terremark.toId(String.valueOf(key.getId())));
            kp.setProviderOwnerId(provider.getContext().getAccountNumber());
            kp.setProviderRegionId(provider.getContext().getRegionId());
            kp.setPublicKey(null);
            kps.add(kp);
        }

        return kps;
	}
    
    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }
}
