/**
 * Copyright (C) 2010-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.terremark.network;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.dc.DataCenter;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.terremark.Terremark;
import org.dasein.cloud.jclouds.terremark.Vdc;
import org.dasein.cloud.network.AddressType;
import org.dasein.cloud.network.IpAddress;
import org.dasein.cloud.network.IpAddressSupport;
import org.dasein.cloud.network.IpForwardingRule;
import org.dasein.cloud.network.Protocol;
import org.jclouds.rest.ResourceNotFoundException;
import org.jclouds.rest.RestContext;
import org.jclouds.trmk.vcloud_0_8.VCloudResponseException;
import org.jclouds.trmk.vcloud_0_8.domain.InternetService;
import org.jclouds.trmk.vcloud_0_8.domain.Node;
import org.jclouds.trmk.vcloud_0_8.domain.PublicIpAddress;
import org.jclouds.trmk.vcloud_0_8.options.AddInternetServiceOptions;
import org.jclouds.trmk.vcloud_0_8.options.AddNodeOptions;
import org.jclouds.trmk.vcloudexpress.TerremarkVCloudExpressAsyncClient;
import org.jclouds.trmk.vcloudexpress.TerremarkVCloudExpressClient;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

public class PublicIp implements IpAddressSupport {
    static private final Logger logger = Logger.getLogger(PublicIp.class);
    
    static @Nonnull URI getPublicAddressUri(@Nonnull RestContext<TerremarkVCloudExpressClient, TerremarkVCloudExpressAsyncClient> ctx, @Nonnull String id) {
        if( id.startsWith("http") ) {
        	int lastSlash = id.lastIndexOf('/');
        	id = id.substring(lastSlash+1);
        }
        return URI.create(ctx.getEndpoint() + "/extensions/v1.6/publicIp/" + id);
        //return URI.create(ctx.getEndpoint() + "/v" + ctx.getApiVersion() + "/extensions/publicIp/" + id);
        //return URI.create(ctx.getEndpoint() + "/extensions/v" + ctx.getApiVersion() + "/publicIp/" + id);   
    }
    
    static URI getPublicAddressUriv16(RestContext<TerremarkVCloudExpressClient, TerremarkVCloudExpressAsyncClient> ctx, String id) {
        if( id.startsWith("http") ) {
        	int lastSlash = id.lastIndexOf('/');
        	id = id.substring(lastSlash+1);
        }
        //return URI.create(ctx.getEndpoint() + "/v" + ctx.getApiVersion() + "/extensions/publicIp/" + id);
        return URI.create(ctx.getEndpoint() + "/extensions/v1.6/publicIp/" + id);   
    }
    
    static URI getInternetServiceUriv16(RestContext<TerremarkVCloudExpressClient, TerremarkVCloudExpressAsyncClient> ctx, String id) {
        if( id.startsWith("http") ) {
        	int lastSlash = id.lastIndexOf('/');
        	id = id.substring(lastSlash+1);
        }
        return URI.create(ctx.getEndpoint() + "/extensions/v1.6/internetService/" + id);
    }
    
    private Terremark provider = null;
    
    PublicIp(@Nonnull Terremark provider) {
        this.provider = provider;
    }
    
    @Override
    public void assign(@Nonnull String addressId, @Nonnull String toServerId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    public void clear(@Nonnull String serverId) throws InternalException, CloudException {
        for( IpAddress ip : listPublicIpPool(false) ) {
            for( IpForwardingRule rule : listRules(ip.getProviderIpAddressId()) ) {
                if( serverId.equals(rule.getServerId()) ) {
                    stopForward(rule.getProviderRuleId());
                }
            }
        }
    }

    @Override
    public @Nonnull String forward(@Nonnull String addressId, int publicPort, @Nonnull Protocol protocol, int privatePort, @Nonnull String onServerId) throws InternalException, CloudException {
        if( logger.isDebugEnabled() ) {
            logger.debug("enter - forward('" + addressId + "'," + publicPort + "," + protocol + "," + privatePort + ",'" + onServerId + "')");
        }
        try {
            org.jclouds.trmk.vcloud_0_8.domain.Protocol proto;
            String privateAddress = null;
            VirtualMachine server;
    
            server = provider.getComputeServices().getVirtualMachineSupport().getVirtualMachine(onServerId);
            String[] addresses = server.getPrivateIpAddresses();
            
            if( addresses != null && addresses.length > 0 ) {
                privateAddress = addresses[0];
            }
            if( logger.isDebugEnabled() ) {
                logger.debug("forward(): privateAddress=" + privateAddress);
            }
            if( privateAddress == null ) {
                logger.error("No private address exists for " + server.getProviderVirtualMachineId() + "/" + server.getName());
                throw new CloudException("No private address exists for " + server.getProviderVirtualMachineId() + "/" + server.getName());
            }
            switch( protocol ) {
                case TCP: proto = org.jclouds.trmk.vcloud_0_8.domain.Protocol.TCP; break;
                case UDP: proto = org.jclouds.trmk.vcloud_0_8.domain.Protocol.UDP; break;
                default: throw new CloudException("Terremark does not support ICMP forwarding.");
            }
            RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
            try {
            	TerremarkVCloudExpressClient client = ctx.getApi();
                Collection<InternetService> services;
                
                try {
                    URI uri = getPublicAddressUriv16(ctx, addressId);
                    
                    if( logger.isInfoEnabled() ) {
                        logger.info("forward(): Looking up internet services for " + addressId + ": " + uri);
                    }
                    services = client.getInternetServicesOnPublicIp(uri);
                }
                catch( VCloudResponseException e ) {
                    logger.error("Could not load services on IP " + addressId);
                    if( logger.isDebugEnabled() ) {
                        e.printStackTrace();
                    }
                    throw new CloudException(e);
                }
                
                InternetService svc = null;
                Node node = null;
        
                for( InternetService service : services ) {
                    if( logger.isDebugEnabled() ) {
                        logger.info("forward(): Checking " + service.getPort() + " against " + publicPort);
                    }
                    if( service.getPort() == publicPort ) {
                        svc = service;
                        break;
                    }
                }
                if( logger.isDebugEnabled() ) {
                    logger.debug("forward(): service=" + svc);
                }
                if( svc == null ) {
                    try { 
                        if( logger.isInfoEnabled() ) {
                            logger.info("forward(): Adding internet service on " + publicPort + " to " + addressId);
                        }
                        svc = client.addInternetServiceToExistingIp(getPublicAddressUri(ctx, addressId), "CUSTOM", proto, publicPort, new AddInternetServiceOptions[0]);
                    }
                    catch( VCloudResponseException e ) {
                        logger.error("forward(): Could not create internet service for IP address " + addressId + ": " + e.getMessage());
                        if( logger.isDebugEnabled() ) {
                            e.printStackTrace();
                        }
                        throw new CloudException(e);
                    }
                }
                else {
                    if( logger.isDebugEnabled() ) {
                        logger.debug("forward(): service.getId()=" + svc.getId());
                    }
                    try {
                        for( Node n : client.getNodes(svc.getId()) ) {
                            if( logger.isDebugEnabled() ) {
                                logger.debug("forward(): Checking private port from node " + n.getPort() + " against " + privatePort);
                            }
                            if( n.getPort() == privatePort ) {
                                node = n;
                                break;
                            }
                        }
                    }
                    catch( RuntimeException e ) {
                        logger.warn("forward(): Could not identify any existing node: " + e.getMessage());
                    }
                }
                if( node != null ) {
                    try {
                        if( logger.isInfoEnabled() ) {
                            logger.info("forward(): Node exists for " + privatePort + " removing, " + node.getId());
                        }
                        client.deleteNode(node.getId());
                    }
                    catch( RuntimeException e ) {
                        logger.warn("forward(): Unable to delete node " + node.getId() + ": " + e.getMessage());
                    }
                }
                try {
                    if( logger.isInfoEnabled() ) {
                        logger.info("forward(): Adding node " + privatePort + " to " + addressId + " against " + svc.getId());
                    }
                    node = client.addNode(svc.getId(), privateAddress, "Forward", privatePort, new AddNodeOptions[0]);
                }
                catch( VCloudResponseException e ) {
                    logger.error("forward(): Failed to add node to internet service: " + e.getMessage());
                    if( logger.isInfoEnabled() ) {
                        e.printStackTrace();
                    }
                    throw new CloudException(e);
                }
                String nodeId = Terremark.toId(String.valueOf(node.getId()));
                
                if( logger.isInfoEnabled() ) {
                    logger.info("forward(): New node is " + nodeId + " (" + node.getId() + ")");
                }
                return nodeId;
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("forward(): Runtime exception while executing method: " + e.getMessage());
            e.printStackTrace();
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("forward(): Error while executing method: " + e.getMessage());
            e.printStackTrace();
            throw new InternalException(e);
        }
        finally {
            logger.debug("exit - forward()");
        }
    }

    @Override
    public @Nullable IpAddress getIpAddress(@Nonnull String addressId) throws InternalException, CloudException {
        for( IpAddress address : listPublicIpPool(false) ) {
            if( address.getProviderIpAddressId().equals(addressId) ) {
                return address;
            }
        }
        return null;
    }

    @Override
    public @Nonnull String getProviderTermForIpAddress(@Nonnull Locale locale) {
        return "IP address";
    }
    
    private @Nullable String getServerAt(@Nonnull String address) throws InternalException, CloudException {
        for( VirtualMachine server : provider.getComputeServices().getVirtualMachineSupport().listVirtualMachines() ) {
            for( String privateAddress : server.getPrivateIpAddresses() ) {
                if( privateAddress.equals(address) ) {
                    return server.getProviderVirtualMachineId();
                }
            }
        }
        return null;
    }

    @Override
    public boolean isAssigned(@Nonnull AddressType type) {
        return false;
    }

    @Override
    public boolean isForwarding() {
        return true;
    }

    @Override
    public boolean isRequestable(@Nonnull AddressType type) {
        return type.equals(AddressType.PUBLIC);
    }
    
    public boolean isSubscribed() {
        return true;
    }
    
    @Override
    public @Nonnull Collection<IpAddress> listPrivateIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    @Override
    public @Nonnull Collection<IpAddress> listPublicIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<IpAddress> matches = new ArrayList<IpAddress>();
            Collection<PublicIpAddress> addresses;
            
            try {
            	TerremarkVCloudExpressClient client = ctx.getApi();
                addresses = client.getPublicIpsAssociatedWithVDC(client.findVDCInOrgNamed(null, null).getHref());
            }
            catch( RuntimeException e ) {
                logger.error(e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( PublicIpAddress addr : addresses ) {
                IpAddress ip = new IpAddress();
                
                ip.setAddress(addr.getAddress());
                ip.setIpAddressId(Terremark.toId(String.valueOf(addr.getId())));
                ip.setRegionId(provider.getContext().getRegionId());
                ip.setServerId(null);
                ip.setProviderLoadBalancerId(null);
                ip.setAddressType(AddressType.PUBLIC);
                matches.add(ip);
            }
            return matches;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public @Nonnull Collection<IpForwardingRule> listRules(@Nonnull String addressId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<IpForwardingRule> rules = new ArrayList<IpForwardingRule>();
            Collection<InternetService> services;
            TerremarkVCloudExpressClient client = ctx.getApi();
            
            try {
                services = client.getInternetServicesOnPublicIp(getPublicAddressUriv16(ctx, addressId));
            }
            catch( RuntimeException e ) {
                logger.error("Could not load services on IP " + addressId);
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( InternetService svc : services ) {
                    IpForwardingRule rule = new IpForwardingRule();
                    
                    logger.debug(String.valueOf(svc.getId()));
                    rule.setAddressId(addressId);
                    rule.setPrivatePort(svc.getPort());
                    rule.setProtocol(toDaseinProtocol(svc.getProtocol()));
                    rule.setProviderRuleId(Terremark.toId(String.valueOf(svc.getId())));
                    rule.setPublicPort(svc.getPort());
                    for( Node node : client.getNodes(svc.getId()) ) {
                    	rule.setServerId(getServerAt(node.getIpAddress()));
                    }  
                    if( logger.isDebugEnabled() ) {
                        logger.debug("* " + rule.getProviderRuleId() + " [" + rule + "]");
                    }
                    rules.add(rule);
            }
            return rules;
        }
        finally {
            ctx.close();
        }
    }   

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public void releaseFromPool(@Nonnull String addressId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            Collection<InternetService> services;
            TerremarkVCloudExpressClient client = ctx.getApi();
            
            try {
                services = client.getInternetServicesOnPublicIp(getPublicAddressUriv16(ctx, addressId)); // TODO: support private
            }
            catch( RuntimeException e ) {
                logger.error("Could not load services on IP " + addressId);
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( InternetService svc : services ) {
                for( Node node : client.getNodes(svc.getId()) ) {
                    if( node != null ) {
                    	client.deleteNode(node.getId());
                    }
                }
                client.deleteInternetService(svc.getId());
            }
            client.deletePublicIp(getPublicAddressUri(ctx, addressId));
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public void releaseFromServer(@Nonnull String addressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();        
    }

    @Override
    public @Nonnull String request(@Nonnull AddressType type) throws InternalException, CloudException {
        if( logger.isDebugEnabled() ) {
            logger.debug("enter - request(" + type + ")");
        }
        try {
            if( type.equals(AddressType.PRIVATE) ) {
                throw new OperationNotSupportedException("Allocation of private IP addresses is not supported in Terremark.");
            }
            RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
            
            try {
                InternetService svc = null;
                
                for( DataCenter dc : provider.getDataCenterServices().listDataCenters(provider.getContext().getRegionId()) ) {
                    try {
                        URI vdcUri = Vdc.getVdcUri(ctx, dc.getProviderDataCenterId());
                        
                        if( logger.isInfoEnabled() ) {
                            logger.info("request(): Requesting " + type + " address in " + dc.getProviderDataCenterId() + " (" + vdcUri + ")");
                        }
                        svc = ctx.getApi().addInternetServiceToVDC(Vdc.getVdcExtensionsUri(ctx, dc.getProviderDataCenterId()), "SSH", org.jclouds.trmk.vcloud_0_8.domain.Protocol.TCP, 22, AddInternetServiceOptions.Builder.withDescription("SVC"));
                    }
                    catch( ResourceNotFoundException e ) {
                        logger.warn("request(): Could not find data center " + dc.getProviderDataCenterId() + ": " + e.getMessage());
                        continue;
                    }
                    catch( VCloudResponseException e ) {
                        logger.error("request(): Failed to execute request for " + type + ": " + e.getMessage());
                        if( logger.isDebugEnabled() ) {
                            e.printStackTrace();
                        }
                        throw new CloudException(e);
                    }
                    break;
                }
                if( svc == null ) {
                    logger.error("request(): No internet service created. Issue with VDC?");
                    throw new CloudException("No internet service created. Issue with VDC?");
                }
                PublicIpAddress address = (svc == null ? null : svc.getPublicIpAddress());
                String id = (address == null ? null : Terremark.toId(String.valueOf(address.getId())));
                
                if( logger.isInfoEnabled() ) {
                    logger.info("request(): New address is " + (address == null ? null : address.getAddress()) + ": " + id + " (" + (address == null ? null : address.getId()) + ")");
                }
                return id;
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("request(): Runtime exception executing request: " + e.getMessage());
            e.printStackTrace();
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("request(): Error executing request: " + e.getMessage());
            e.printStackTrace();
            throw new InternalException(e);
        }
        finally {
            logger.debug("exit - request()");
        }
    }
    
    @Override
    public void stopForward(@Nonnull String ruleId) throws InternalException, CloudException {

        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        URI uri = getInternetServiceUriv16(ctx, ruleId);
    	try {
    		TerremarkVCloudExpressClient client = ctx.getApi();
    		for( Node node : client.getNodes(uri) ) {

    			client.deleteNode(node.getId()); // TODO: fix me
    		}
    		client.deleteInternetService(uri);
        }
        catch( ResourceNotFoundException e ) {
            logger.warn("request(): Could not find forwarding rule " + ruleId + ": " + e.getMessage());
            throw new InternalException(e);
        }
    	
        finally {
            ctx.close();
        }
    }
    
    private @Nonnull Protocol toDaseinProtocol(@Nonnull org.jclouds.trmk.vcloud_0_8.domain.Protocol jcloudsProtocol) {
        switch( jcloudsProtocol ) {
            case TCP: case HTTP: case HTTPS: case FTP: return Protocol.TCP;
            case UDP: return Protocol.UDP;
            default: return Protocol.TCP;
        }
    }
}
